(*
 * Copyright (C) 2013 National ICT Australia Ltd ('NICTA')
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Author: David Cock - David.Cock@nicta.com.au *)

(* XXX - this needs more explanation *)

header "Induction on sound predicates"

theory Expectation_Induction imports Expectations Transformers begin

section {* Fixed points for functions with bounded iterates. *}

text {* Fixed points restricted to a set *}

definition
  lfp_in :: "('a::complete \<Rightarrow> 'a) \<Rightarrow> 'a set \<Rightarrow> 'a"
where
  "lfp_in f S \<equiv> if \<exists>x\<in>S. f x \<le> x then glb {x\<in>S. f x \<le> x} else upperbound S"

lemma lfp_in_lowerbound:
  "\<lbrakk> bounded_below S; x \<in> S; f x \<le> x \<rbrakk> \<Longrightarrow> lfp_in f S \<le> x"
  unfolding lfp_in_def by(clarsimp, blast intro!:glb_lower)

lemma lfp_in_greatest:
  "\<lbrakk> bounded_above S; z \<in> S; \<And>x. \<lbrakk> x \<in> S; f x \<le> x \<rbrakk> \<Longrightarrow> z \<le> x \<rbrakk> \<Longrightarrow> z \<le> lfp_in f S"
  unfolding lfp_in_def by(clarsimp, blast intro:glb_greatest)

definition
  gfp_in :: "('a::complete \<Rightarrow> 'a) \<Rightarrow> 'a set \<Rightarrow> 'a"
where
  "gfp_in f S \<equiv> if \<exists>x\<in>S. x \<le> f x then lub {x\<in>S. x \<le> f x} else lowerbound S"

lemma gfp_in_upperbound:
  "\<lbrakk> bounded_above S; x \<in> S; x \<le> f x \<rbrakk> \<Longrightarrow> x \<le> gfp_in f S"
  unfolding gfp_in_def
  by(clarsimp, blast intro!:lub_greater dest!:bounded_aboveD)

lemma gfp_in_least:
  "\<lbrakk> bounded_below S; z \<in> S; \<And>x. \<lbrakk> x \<in> S; x \<le> f x \<rbrakk> \<Longrightarrow> x \<le> z \<rbrakk> \<Longrightarrow> gfp_in f S \<le> z"
  unfolding gfp_in_def by(clarsimp, blast intro:lub_least)

definition
  closed_under :: "('a \<Rightarrow> 'a) \<Rightarrow> 'a set \<Rightarrow> bool"
where
  "closed_under f S \<equiv> \<forall>s\<in>S. f s \<in> S"

lemma closed_underI[intro]:
  "\<lbrakk> \<And>x. x \<in> S \<Longrightarrow> f x \<in> S \<rbrakk> \<Longrightarrow> closed_under f S"
  unfolding closed_under_def by(simp)

lemma closed_underD[dest]:
  "\<lbrakk> closed_under f S; s \<in> S \<rbrakk> \<Longrightarrow> f s \<in> S"
  by(simp add:closed_under_def)

lemma lfp_in_closed[intro]:
  assumes b_bl: "bounded_below S"
      and b_ab: "bounded_above S"
      and dn_c: "down_complete S"
      and cl_u: "closed_under f S"
  shows "lfp_in f S \<in> S"
proof -
  from b_ab obtain t where tin: "t \<in> S" and ttop: "\<forall>x\<in>S. x \<le> t"
    by(blast)
  from b_ab and cl_u obtain x where xin: "x \<in> S" and xfp: "f x \<le> x"
    by(blast)
  hence "\<exists>x\<in>S. f x \<le> x" by(blast)
  thus ?thesis
    unfolding lfp_in_def
    using b_bl and dn_c and cl_u and xin and xfp
      and tin and ttop
    by(simp, blast intro:down_completeD)
qed

lemma gfp_in_closed[intro]:
  assumes b_bl: "bounded_below S"
      and b_ab: "bounded_above S"
      and up_c: "up_complete S"
      and cl_u: "closed_under f S"
  shows "gfp_in f S \<in> S"
proof -
  from b_ab obtain t where tin: "t \<in> S" and ttop: "\<forall>x\<in>S. x \<le> t"
    by(blast)
  from b_bl and cl_u obtain x where xin: "x \<in> S" and xfp: "x \<le> f x"
    by(blast)
  hence "\<exists>x\<in>S. x \<le> f x" by(blast)
  thus ?thesis
    unfolding gfp_in_def
    using up_c and cl_u and xin and xfp
      and tin and ttop
    by(simp, blast intro:up_completeD)
qed

lemma healthy_closed[simp,intro]:
  "healthy t \<Longrightarrow> closed_under t sound_expectations"
  by(blast)

text {* Monotonicity on a subset *}

definition
  mono_on :: "('a::ord \<Rightarrow> 'a) \<Rightarrow> 'a set \<Rightarrow> bool"
where
  "mono_on f S \<equiv> \<forall>x\<in>S. \<forall>y\<in>S. x \<le> y \<longrightarrow> f x \<le> f y"

lemma mono_onI[intro]:
  "\<lbrakk> \<And>x y. \<lbrakk> x \<in> S; y \<in> S; x \<le> y \<rbrakk> \<Longrightarrow> f x \<le> f y \<rbrakk> \<Longrightarrow> mono_on f S"
  by(simp add:mono_on_def)

lemma mono_onD[dest]:
  "\<lbrakk> mono_on f S; x \<in> S; y \<in> S; x \<le> y \<rbrakk> \<Longrightarrow> f x \<le> f y"
  by(simp add:mono_on_def)

lemma lfp_in_lemma2:
  "\<lbrakk> bounded_below S; bounded_above S;
     down_complete S; closed_under f S; mono_on f S \<rbrakk> \<Longrightarrow>
   f (lfp_in f S) \<le> lfp_in f S"
  apply(rule lfp_in_greatest, assumption+)
   apply(blast intro:closed_underD[OF _ lfp_in_closed])
  apply(blast intro:lfp_in_closed lfp_in_lowerbound order_trans)
  done

lemma gfp_in_lemma2:
  "\<lbrakk> bounded_below S; bounded_above S;
     up_complete S; closed_under f S; mono_on f S \<rbrakk> \<Longrightarrow>
   gfp_in f S \<le> f (gfp_in f S)"
  apply(rule gfp_in_least, assumption+)
   apply(blast intro:closed_underD[OF _ gfp_in_closed])
  apply(blast intro:gfp_in_closed gfp_in_upperbound order_trans)
  done

lemma lfp_in_lemma3:
  "\<lbrakk> bounded_below S; bounded_above S;
     down_complete S; closed_under f S; mono_on f S \<rbrakk> \<Longrightarrow>
   lfp_in f S \<le> f (lfp_in f S)"
  by(blast intro!:lfp_in_closed lfp_in_lemma2 lfp_in_lowerbound
           dest:closed_underD)

lemma gfp_in_lemma3:
  "\<lbrakk> bounded_below S; bounded_above S;
     up_complete S; closed_under f S; mono_on f S \<rbrakk> \<Longrightarrow>
   f (gfp_in f S) \<le> gfp_in f S"
  by(blast intro!:gfp_in_closed gfp_in_lemma2 gfp_in_upperbound
           dest:closed_underD)

lemma lfp_in_unfold:
  "\<lbrakk> bounded_below S; bounded_above S;
     down_complete S; closed_under f S; mono_on f S \<rbrakk> \<Longrightarrow>
   lfp_in f S = f (lfp_in f S)"
  by (iprover intro: order_antisym lfp_in_lemma2 lfp_in_lemma3)

lemma gfp_in_unfold:
  "\<lbrakk> bounded_below S; bounded_above S;
     up_complete S; closed_under f S; mono_on f S \<rbrakk> \<Longrightarrow>
   gfp_in f S = f (gfp_in f S)"
  by (iprover intro: order_antisym gfp_in_lemma2 gfp_in_lemma3)

lemma lfp_in_mono:
  assumes subset: "S \<subseteq> T"
      and bounded_below_S: "bounded_below S"
      and bounded_above_S: "bounded_above S"
      and bounded_below_T: "bounded_below T"
      and down_complete_S: "down_complete S"
      and closed_under_S:  "closed_under f S"
      and mono_on_S:       "mono_on f S"
  shows "lfp_in f T \<le> lfp_in f S"
proof -
  from assms have "lfp_in f S \<in> T" by(blast intro:lfp_in_closed)
  with assms show ?thesis
    by(blast intro:lfp_in_lowerbound lfp_in_lemma2)
qed

lemma gfp_in_mono:
  assumes subset: "S \<subseteq> T"
      and bounded_below_S: "bounded_below S"
      and bounded_above_S: "bounded_above S"
      and bounded_above_T: "bounded_above T"
      and up_complete_S:   "up_complete S"
      and closed_under_S:  "closed_under f S"
      and mono_on_S:       "mono_on f S"
  shows "gfp_in f S \<le> gfp_in f T"
proof -
  from assms have "gfp_in f S \<in> T" by(blast intro:gfp_in_closed)
  with assms show ?thesis
    by(blast intro:gfp_in_upperbound gfp_in_lemma2)
qed

lemma lfp_in_ordinal_induct:
  assumes bounded_below: "bounded_below S"
      and bounded_above: "bounded_above S"
      and up_complete:   "up_complete S"
      and down_complete: "down_complete S"
      and closed:        "closed_under f S"
      and mono:          "mono_on f S"
      and P_preserved:   "\<And>s. \<lbrakk> s \<in> S; P s \<rbrakk> \<Longrightarrow> P (f s)"
      and P_lub:         "\<And>M t. \<lbrakk> \<exists>s. s \<in> M; M \<subseteq> S; \<forall>s\<in>M. P s \<rbrakk> \<Longrightarrow> P (lub M)"
      and P_lowerbound:  "P (lowerbound S)"
  shows "P (lfp_in f S)"
proof -
  from assms have lfp_closed: "lfp_in f S \<in> S"
    by(blast intro: lfp_in_closed)

  from bounded_below and bounded_above
   and down_complete and closed and mono
  have lfp_fp: "f (lfp_in f S) \<le> lfp_in f S" by(rule lfp_in_lemma2)
  hence bex_fp: "\<exists>s\<in>S. f s \<le> s" by(blast intro:lfp_closed)
  
  from bex_fp have lfp_simp:
    "lfp_in f S = glb {x \<in> S. f x \<le> x}"
    unfolding lfp_in_def by(simp)

  let ?M = "{s\<in>S. s \<le> lfp_in f S \<and> P s}"

  from bounded_below and lfp_closed and P_lowerbound
  have P_lub_M: "P (lub ?M)" by(blast intro:P_lub)

  moreover have "lub ?M = lfp_in f S"
    proof(rule antisym)
      from up_complete and bounded_above
      have X: "lub ?M \<in> S"
        using P_lowerbound and bounded_below and lfp_closed
        by(blast intro:up_completeD)

      (* XXX - Tidy this up *)

      with bex_fp and bounded_below show "lfp_in f S \<le> lub ?M"
        apply -
        apply(rule lfp_in_lowerbound, assumption+)
        apply(rule lub_greater)
         apply(safe)
           apply(blast intro:closed_underD[OF closed])
          apply(rule order_trans[OF _ lfp_fp])
          apply(rule_tac f=f in mono_onD)
             apply(rule mono)
            apply(assumption)
           apply(rule lfp_closed)
          using P_lowerbound and lfp_closed apply(blast intro:lub_least)
         apply(rule P_preserved, simp)
         apply(rule P_lub_M)
        apply(blast)
        done

      from bounded_below and bounded_above show "lub ?M \<le> lfp_in f S"
        by(blast intro:order_trans lfp_in_lowerbound lub_least
                       lfp_in_greatest X P_lowerbound)
    qed

  ultimately show ?thesis by(simp)
qed

lemma gfp_in_ordinal_induct:
  assumes bounded_below: "bounded_below S"
      and bounded_above: "bounded_above S"
      and up_complete:   "up_complete S"
      and down_complete: "down_complete S"
      and closed:        "closed_under f S"
      and mono:          "mono_on f S"
      and P_preserved:   "\<And>s. \<lbrakk> s \<in> S; P s \<rbrakk> \<Longrightarrow> P (f s)"
      and P_glb:         "\<And>M t. \<lbrakk> \<exists>s. s \<in> M; M \<subseteq> S; \<forall>s\<in>M. P s \<rbrakk> \<Longrightarrow> P (glb M)"
      and P_upperbound:  "P (upperbound S)"
  shows "P (gfp_in f S)"
proof -
  from assms have gfp_closed: "gfp_in f S \<in> S"
    by(blast intro: gfp_in_closed)

  from bounded_below and bounded_above
   and up_complete and closed and mono
  have gfp_fp: "gfp_in f S \<le> f (gfp_in f S)" by(rule gfp_in_lemma2)
  hence bex_fp: "\<exists>s\<in>S. s \<le> f s" by(blast intro:gfp_closed)
  
  from bex_fp have gfp_simp:
    "gfp_in f S = lub {x \<in> S. x \<le> f x}"
    unfolding gfp_in_def by(simp)

  let ?M = "{s\<in>S. gfp_in f S \<le> s \<and> P s}"

  from bounded_above and gfp_closed and P_upperbound
  have P_glb_M: "P (glb ?M)" by(blast intro:P_glb)

  moreover have "glb ?M = gfp_in f S"
    proof(rule antisym)
      from down_complete and bounded_below
      have X: "glb {s \<in> S. gfp_in f S \<le> s \<and> P s} \<in> S"
        using P_upperbound and bounded_above and gfp_closed
        by(blast intro:down_completeD)

      (* XXX - tidy *)

      with bex_fp and bounded_above show "glb ?M \<le> gfp_in f S"
        apply -
        apply(rule gfp_in_upperbound, assumption+)
        apply(rule glb_lower)
         apply(safe)
           apply(blast intro:closed_underD[OF closed])
          apply(rule order_trans[OF gfp_fp])
          apply(rule_tac f=f in mono_onD)
             apply(rule mono)
            apply(rule gfp_closed)
           apply(assumption)
          using P_upperbound and gfp_closed apply(blast intro:glb_greatest)
         apply(rule P_preserved, simp)
        apply(rule P_glb_M)
       apply(blast)
       done

      from bounded_below and bounded_above show "gfp_in f S \<le> glb ?M"
        by(blast intro:order_trans gfp_in_upperbound glb_greatest
                       gfp_in_least X P_upperbound)
    qed

  ultimately show ?thesis by(simp)
qed

end
