header "Probabilistic Correspondence for pGCL"

theory PCorres imports pGCL DataRefinement begin

text {* @{term pcorres}, or probabilistic correspondence is defined by
        equality on distribution transformers, modulo a guard. *}
definition
  pcorres :: "('b \<Rightarrow> 'a) \<Rightarrow> ('b \<Rightarrow> bool) \<Rightarrow> 'a prog \<Rightarrow> 'b prog \<Rightarrow> bool"
where
  "pcorres \<phi> G A B \<longleftrightarrow>
   (\<forall>Q. (sound Q \<and> bounded_by 1 Q) \<longrightarrow>
        \<guillemotleft>G\<guillemotright> && (wp A Q o \<phi>) = \<guillemotleft>G\<guillemotright> && wp B (Q o \<phi>))"

lemma pcorresI:
  "\<lbrakk> \<And>Q. \<lbrakk> sound Q; bounded_by 1 Q \<rbrakk> \<Longrightarrow>
         \<guillemotleft>G\<guillemotright> && (wp A Q o \<phi>) = \<guillemotleft>G\<guillemotright> && wp B (Q o \<phi>) \<rbrakk> \<Longrightarrow>
   pcorres \<phi> G A B"
  by(simp add:pcorres_def)

text {* Often easier to use, as it allows one to assume the precondition. *}
lemma pcorresI2[intro]:
  fixes A::"'a prog" and B::"'b prog"
  assumes withG: "\<And>Q s. \<lbrakk> sound Q; bounded_by 1 Q; G s \<rbrakk> \<Longrightarrow>
                        wp A Q (\<phi> s)= wp B (Q o \<phi>) s"
      and hA: "healthy (wp A)"
      and hB: "healthy (wp B)"
  shows "pcorres \<phi> G A B"
proof(rule pcorresI, rule ext)
  fix Q::"'a \<Rightarrow> real" and s::'b
  assume sQ: "sound Q" and bQ: "bounded_by 1 Q"
  show "(\<guillemotleft>G\<guillemotright> && (wp A Q \<circ> \<phi>)) s = (\<guillemotleft>G\<guillemotright> && wp B (Q \<circ> \<phi>)) s"
  proof(cases "G s")
    case True note this
    moreover
    from hA and sQ and bQ have "0 \<le> wp A Q (\<phi> s)" by(blast)
    moreover
    from hB and sQ and bQ have "0 \<le> wp B (Q o \<phi>) s" by(blast)
    ultimately show ?thesis
      using sQ and bQ by(simp add:exp_conj_def withG)
  next
    case False note this
    moreover
    from hA and sQ and bQ have "wp A Q (\<phi> s) \<le> 1" by(blast)
    moreover
    from hB and sQ and bQ have "wp B (Q o \<phi>) s \<le> 1"
      by(blast dest:healthy_bounded_byD intro:sound_nneg)
    ultimately show ?thesis by(simp add:exp_conj_def)
  qed
qed

lemma pcorresD:
  "\<lbrakk> pcorres \<phi> G A B; sound Q; bounded_by 1 Q \<rbrakk> \<Longrightarrow>
   \<guillemotleft>G\<guillemotright> && (wp A Q o \<phi>) = \<guillemotleft>G\<guillemotright> && wp B (Q o \<phi>)"
  unfolding pcorres_def by(simp)

text {* Again, easier to use if the precondition is known to hold. *}
lemma pcorresD2:
  "\<lbrakk> pcorres \<phi> G A B; sound Q; bounded_by 1 Q;
     healthy (wp A); healthy (wp B); G s \<rbrakk> \<Longrightarrow>
   wp A Q (\<phi> s) = wp B (Q o \<phi>) s"
  apply(drule pcorresD, assumption+)
  apply(drule_tac x=s in fun_cong)
  apply(subgoal_tac "0 \<le> wp A Q (\<phi> s)")
   apply(subgoal_tac "0 \<le> wp B (Q o \<phi>) s")
    apply(simp add:exp_conj_def)
   apply(rule_tac x=s in nnegD)
   apply(rule_tac t="wp B" in healthy_nnegD, blast, blast)
  apply(blast)
  done

text {* Correspondence composes over sequential composition. *}
lemma pcorres_Seq:
  fixes A::"'b prog" and B::"'c prog"
    and C::"'b prog" and D::"'c prog"
    and \<phi>::"'c \<Rightarrow> 'b"
  assumes pcAB: "pcorres \<phi> G A B"
      and pcCD: "pcorres \<phi> H C D"
      and hA: "healthy (wp A)" and hB: "healthy (wp B)"
      and hC: "healthy (wp C)" and hD: "healthy (wp D)"
      and p3p2: "\<And>Q. \<lbrakk> sound Q; bounded_by 1 Q \<rbrakk> \<Longrightarrow>
                     \<guillemotleft>I\<guillemotright> && wp B Q = wp B (\<guillemotleft>H\<guillemotright> && Q)"
      and p1p3: "\<And>s. G s \<Longrightarrow> I s"
  shows "pcorres \<phi> G (A;;C) (B;;D)"
proof(rule pcorresI)
  fix Q::"'b \<Rightarrow> real"
  assume sQ: "sound Q" and bQ: "bounded_by 1 Q"
  with hC have sCQ: "sound (wp C Q)" and bCQ: "bounded_by 1 (wp C Q)"
    by(blast+)
  from sQ and bQ and hD have sDQ: "sound (wp D (Q o \<phi>))"
                         and bDQ: "bounded_by 1 (wp D (Q o \<phi>))"
    by(blast intro:sound_nneg)+

  have p3p1: "\<And>R S. \<lbrakk> sound R; sound S;
                      bounded_by 1 R; bounded_by 1 S;
                      \<guillemotleft>I\<guillemotright> && R = \<guillemotleft>I\<guillemotright> && S \<rbrakk> \<Longrightarrow>
                    \<guillemotleft>G\<guillemotright> && R = \<guillemotleft>G\<guillemotright> && S"
  proof(rule ext)
    fix R::"'c \<Rightarrow> real" and S::"'c \<Rightarrow> real" and s::'c
    assume a3: "\<guillemotleft>I\<guillemotright> && R = \<guillemotleft>I\<guillemotright> && S"
       and sR: "sound R" and sS: "sound S"
       and bR: "bounded_by 1 R" and bS: "bounded_by 1 S"
    show "(\<guillemotleft>G\<guillemotright> && R) s = (\<guillemotleft>G\<guillemotright> && S) s"
    proof(simp add:exp_conj_def, cases "G s")
      case False note this
      moreover from bR have "R s \<le> 1" by(blast)
      moreover from bS have "S s \<le> 1" by(blast)
      ultimately show "\<guillemotleft>G\<guillemotright> s .& R s = \<guillemotleft>G\<guillemotright> s .& S s"
        by(simp)
    next
      case True note p1 = this
      with p1p3 have "I s" by(blast)
      with fun_cong[OF a3, where x=s] have "1 .& R s = 1 .& S s"
        by(simp add:exp_conj_def)
      with p1 show "\<guillemotleft>G\<guillemotright> s .& R s = \<guillemotleft>G\<guillemotright> s .& S s"
        by(simp)
    qed
  qed

  show "\<guillemotleft>G\<guillemotright> && (wp (A;;C) Q o \<phi>) = \<guillemotleft>G\<guillemotright> && wp (B;;D) (Q o \<phi>)"
    apply(simp add:wp_eval)
    apply(simp add:pcorresD[OF pcAB] sCQ bCQ)
    apply(rule p3p1)
        apply(blast intro:sQ hB hC)
       apply(blast intro:sQ hB hD)
      apply(rule_tac t="wp B" in healthy_bounded_byD)
        apply(rule hB)
       apply(blast intro:bCQ)
      apply(blast intro:sCQ sound_nneg)
     apply(rule_tac t="wp B" in healthy_bounded_byD)
       apply(rule hB)
      apply(blast intro:bDQ)
     apply(blast intro:sDQ sound_nneg)
    apply(simp add:sCQ bCQ sDQ bDQ p3p2)
    apply(rule_tac f="wp B" in arg_cong)
    apply(simp add:pcorresD[OF pcCD] sQ bQ)
    done
qed

lemma pcorres_Skip:
  "pcorres \<phi> G Skip Skip"
  by(simp add:pcorres_def wp_eval)

text {* Correspondence implies (data) refinement *}
lemma pcorres_drefine:
  assumes corres: "pcorres \<phi> G A C"
      and hC: "healthy (wp C)"
  shows "drefines \<phi> G A C"
proof
  fix P and Q
  assume sP: "sound P" and sQ: "sound Q"
     and bP: "bounded_by 1 P" and bQ: "bounded_by 1 Q"
     and wpA: "P \<tturnstile> wp A Q"
  from wpA have "P o \<phi> \<tturnstile> wp A Q o \<phi>" by(simp add:o_def le_fun_def)
  hence "\<guillemotleft>G\<guillemotright> && (P o \<phi>) \<tturnstile> \<guillemotleft>G\<guillemotright> && (wp A Q o \<phi>)"
    by(rule exp_conj_mono_right)
  also from corres and sQ and bQ
  have "... = \<guillemotleft>G\<guillemotright> && (wp C (Q o \<phi>))" by(rule pcorresD)
  also
  have "... \<tturnstile> wp C (Q o \<phi>)"
  proof(rule le_funI)
    fix s
    from hC and sQ have nn_wpC: "0 \<le> wp C (Q o \<phi>) s" by(blast)
    show "(\<guillemotleft>G\<guillemotright> && wp C (Q o \<phi>)) s \<le> wp C (Q o \<phi>) s"
    proof(cases "G s")
      case True
      with nn_wpC show ?thesis by(simp add:exp_conj_def)
    next
      case False note this
      moreover {
        from sQ have "sound (Q o \<phi>)" by(simp)
        moreover from bQ have "bounded_by 1 (Q o \<phi>)" by(simp)
        moreover note hC
        ultimately have "wp C (Q o \<phi>) s \<le> 1" by(blast)
      }
      moreover note nn_wpC
      ultimately show ?thesis by(simp add:exp_conj_def)
    qed
  qed
  finally show "\<guillemotleft>G\<guillemotright> && (P o \<phi>) \<tturnstile> wp C (Q o \<phi>)" .
qed

lemma refines_determ:
  fixes a::"'a prog" and b::"'b prog"
  assumes da: "determ (wp a)"
      and ha: "healthy (wp a)"
      and hb: "healthy (wp b)"
      and dr: "drefines \<phi> G a b"
      and sab: "sub_add (wp b)"
  shows "pcorres \<phi> G a b"
proof(rule pcorresI, rule contrapos_pp)
  from hb show "healthy (wp b)" .

  fix Q::"'a \<Rightarrow> real"
  assume sQ: "sound Q" and bQ: "bounded_by 1 Q"
  and ne: "\<guillemotleft>G\<guillemotright> && (wp a Q o \<phi>) \<noteq> \<guillemotleft>G\<guillemotright> && wp b (Q o \<phi>)"

  from ne have ne': "wp a Q o \<phi> \<noteq> wp b (Q o \<phi>)"
    by(rule contrapos_nn, simp)

  from ha and sQ and bQ
  have "\<guillemotleft>G\<guillemotright> && (wp a Q o \<phi>) \<tturnstile> wp b (Q o \<phi>)" by(blast intro!:drefinesD[OF dr])
  hence le: "\<guillemotleft>G\<guillemotright> && (wp a Q o \<phi>) \<tturnstile> \<guillemotleft>G\<guillemotright> && wp b (Q o \<phi>)"
    apply(simp add:exp_conj_def)
    apply(rule le_funI)
    apply(erule_tac x=x in le_funE)
    apply(case_tac "G x", simp_all)
     apply(subgoal_tac "0 \<le> wp a Q (\<phi> x)", simp)
     using ha and sQ apply(blast)
    apply(subgoal_tac "wp a Q (\<phi> x) \<le> 1", simp)
    using ha and sQ and bQ apply(blast)
    done

  have nle: "\<exists>s. (\<guillemotleft>G\<guillemotright> && (wp a Q o \<phi>)) s < (\<guillemotleft>G\<guillemotright> && wp b (Q o \<phi>)) s"
  proof(rule contrapos_np[OF ne], rule ext, rule antisym)
    fix s
    from le show "(\<guillemotleft>G\<guillemotright> && (wp a Q o \<phi>)) s \<le> (\<guillemotleft>G\<guillemotright> && wp b (Q o \<phi>)) s"
      by(blast)
  next
    fix s
    assume "\<not> (\<exists>s. (\<guillemotleft>G\<guillemotright> && (wp a Q \<circ> \<phi>)) s < (\<guillemotleft>G\<guillemotright> && wp b (Q \<circ> \<phi>)) s)"
    thus " (\<guillemotleft>G\<guillemotright> && (wp b (Q \<circ> \<phi>))) s \<le> (\<guillemotleft>G\<guillemotright> && (wp a Q \<circ> \<phi>)) s"
      by(simp add:not_less)
  qed
  from this obtain s where less_s:
    "(\<guillemotleft>G\<guillemotright> && (wp a Q \<circ> \<phi>)) s < (\<guillemotleft>G\<guillemotright> && wp b (Q \<circ> \<phi>)) s"
    by(blast)
  hence larger: "wp a Q (\<phi> s) < wp b (Q \<circ> \<phi>) s"
  proof(cases "G s")
    case True
    moreover from ha and sQ have "0 \<le> wp a Q (\<phi> s)"
      by(blast)
    moreover from hb and sQ have "0 \<le> wp b (Q o \<phi>) s"
      by(blast)
    moreover note less_s
    ultimately show ?thesis by(simp add:exp_conj_def)
  next
    case False
    moreover from ha and sQ and bQ have "wp a Q (\<phi> s) \<le> 1"
      by(blast)
    moreover {
      from bQ have "bounded_by 1 (Q o \<phi>)"
        by(blast)
      with hb and sQ have "wp b (Q o \<phi>) s \<le> 1"
        by(blast intro!:sound_nneg healthy_bounded_byD)
    }
    moreover note less_s
    ultimately show ?thesis by(simp add:exp_conj_def)
  qed
  from less_s have "(\<guillemotleft>G\<guillemotright> && (wp a Q \<circ> \<phi>)) s \<noteq> (\<guillemotleft>G\<guillemotright> && wp b (Q \<circ> \<phi>)) s"
    by(force)
  hence G_s: "G s"
  proof(rule contrapos_np)
    assume nG: "\<not> G s"
    moreover from ha and sQ and bQ have "wp a Q (\<phi> s) \<le> 1"
      by(blast)
    moreover {
      from bQ have "bounded_by 1 (Q o \<phi>)"
        by(blast)
      with hb and sQ have "wp b (Q o \<phi>) s \<le> 1"
        by(blast intro!:sound_nneg healthy_bounded_byD)
    }
    ultimately
    show "(\<guillemotleft>G\<guillemotright> && (wp a Q \<circ> \<phi>)) s = (\<guillemotleft>G\<guillemotright> && wp b (Q \<circ> \<phi>)) s"
      by(simp add:exp_conj_def)
  qed

  let ?Qc = "\<lambda>s. bound_of Q - Q s"

  have sQc: "sound ?Qc" and bQc: "bounded_by 1 ?Qc"
  proof
    have "bound_of Q \<le> 1"
      using bQ by(force intro:bound_of_leI)
    also {
      from sQ have "\<And>s. 0 \<le> Q s" by(blast)
      hence "\<And>s. 1 \<le> 1 + Q s" by(simp)
    }
    finally have "\<And>s. bound_of Q - Q s \<le> 1" by(simp add:sign_simps)
    thus "bounded_by 1 ?Qc" by(blast)
    thus "bounded_by 1 ?Qc" .

    from sQ have "sound Q" by(simp)
    hence "\<And>s. Q s \<le> bound_of Q" by(blast)
    thus "nneg ?Qc"
      apply -
      apply(rule nnegI)
      apply(simp add:sign_simps)
      done
  qed

  from da have "maximal (wp a)" by(blast)
  hence "bound_of Q = wp a (\<lambda>s. bound_of Q) (\<phi> s)"
    by(simp add:maximalD)
  also have "wp a (\<lambda>s. bound_of Q) (\<phi> s) = wp a (\<lambda>s. Q s + ?Qc s) (\<phi> s)"
    by(simp)
  also {
    from da have "additive (wp a)" by(blast)
    with sQ and sQc
    have "wp a (\<lambda>s. Q s + ?Qc s) (\<phi> s) =
          wp a Q (\<phi> s) + wp a ?Qc (\<phi> s)" by(subst additiveD, blast+)
  }
  also {
    from ha and sQc and bQc
    have "\<guillemotleft>G\<guillemotright> && (wp a ?Qc o \<phi>) \<tturnstile> wp b (?Qc o \<phi>)"
      by(blast intro!:drefinesD[OF dr])
    hence "(\<guillemotleft>G\<guillemotright> && (wp a ?Qc o \<phi>)) s \<le> wp b (?Qc o \<phi>) s"
      by(blast)
    moreover from sQc and ha
    have "0 \<le> wp a (\<lambda>s. bound_of Q - Q s) (\<phi> s)"
      by(blast)
    ultimately
    have "wp a ?Qc (\<phi> s) \<le> wp b (?Qc o \<phi>) s"
      using G_s by(simp add:exp_conj_def)
    hence "wp a Q (\<phi> s) + wp a ?Qc (\<phi> s) \<le> wp a Q (\<phi> s) + wp b (?Qc o \<phi>) s"
      by(rule add_left_mono)
    also with larger
    have "wp a Q (\<phi> s) + wp b (?Qc o \<phi>) s <
          wp b (Q o \<phi>) s + wp b (?Qc o \<phi>) s"
      by(auto)
    finally
    have "wp a Q (\<phi> s) + wp a ?Qc (\<phi> s) <
          wp b (Q o \<phi>) s + wp b (?Qc o \<phi>) s" .
  }
  also from sab and sQ and sQc
  have "wp b (Q o \<phi>) s + wp b (?Qc o \<phi>) s \<le>
        wp b (\<lambda>s. (Q o \<phi>) s + (?Qc o \<phi>) s) s"
    by(blast)
  also have "... = wp b (\<lambda>s. bound_of Q) s"
    by(simp)
  finally
  have "\<not> feasible (wp b)"
  proof(rule contrapos_pn)
    assume fb: "feasible (wp b)"
    have "bounded_by (bound_of Q) (\<lambda>s. bound_of Q)" by(blast)
    hence "bounded_by (bound_of Q) (wp b (\<lambda>s. bound_of Q))"
      using sQ by(blast intro:feasible_boundedD[OF fb])
    hence "wp b (\<lambda>s. bound_of Q) s \<le> bound_of Q" by(blast)
    thus "\<not> bound_of Q < wp b (\<lambda>s. bound_of Q) s" by(simp)
  qed
  thus "\<not> healthy (wp b)"
    by(blast intro:healthy_feasibleD)
qed

end
